class BaseCommand():

    def __init__(self, sli):
        self.sli = sli

    def execute(self):
        """SLI starts here to call the run function"""
        self.run()

    def run(self):
        """ Should be overridden by child classes"""
        print('This module has not been implemented')


from .base import BaseCommand
from sli.tools import get_variable_input
from sli.decorators import (
    require_ngfw_connection_params,
    require_single_skillet,
    require_skillet_type
)

class ConfigureCommand(BaseCommand):

    sli_command = 'configure'

    @require_single_skillet
    @require_skillet_type('panos')
    @require_ngfw_connection_params
    def run(self):
        for var in self.sli.skillet.variables:
            self.sli.context.update(get_variable_input(var))
        exe = self.sli.skillet.execute(self.sli.context) 
        if self.sli.commit:
            print('Commiting configuration...')
            self.sli.skillet.panoply.commit()
            print('Finished')
        else:
            print('Configuration loaded into candidate config')


from .base import BaseCommand
from sli.tools import print_table

"""
Load and display a list of skillets in a directory.
Display any errors
"""

class LoadCommand(BaseCommand):

    sli_command = 'load'

    def run(self):
        """SLI action, test load all skillets in directory and print out loaded skillets"""
        objs = []
        for s in self.sli.skillets:
            obj = {
                'name': s.name,
                'type': s.type,
            }
            objs.append(obj)
        print_table(
            objs,
            {
                "Name": "name",
                "Type": "type",
            }
        )


from skilletlib import SkilletLoader
from sli.tools import print_table
from getpass import getpass
from panforge import Report
import os

from sli import commands
from types import ModuleType

class SkilletLineInterface():

    def __init__(self, options, action):
        self.action = action
        self.options = options
        self._unpack_options()
        self.command_map = {}
        self._load_commands()
        self._verify_command()
        self.sl = SkilletLoader()
        self._load_skillets()
        self._verify_loaded_skillets()
        self.context = {}
        self.skillet = None # Active running skillet
    
    def _unpack_options(self):
        """Unpack options onto self where required"""
        self.verbose = self.options.get('verbose', False)
        self.commit = self.options.get('commit', False)
        self.generate_report = self.options.get('report', False)
        self.report_file = self.options.get('report_file', '')

        # If a report file was specified, assume we want to create it
        if self.report_file:
            self.generate_report = True
    
    def _load_commands(self):
        """
        Loads a list of commands from the sli.commands package. Does so by iterating over contents of the
        package and locating types of 'type' that have an sli_command specified.
        """

        # Go through each module in the commands package
        for item in dir(commands):
            item_obj = getattr(commands, item)
            if isinstance(item_obj, ModuleType):

                # Get only the attributes which are classes
                for item_attr in dir(item_obj):
                    item_attr_obj = getattr(item_obj, item_attr)
                    if isinstance(item_attr_obj, type):

                        # Load only SLI command modules that are appropriately written
                        command_string = getattr(item_attr_obj, 'sli_command', '')
                        if len(command_string) > 0:
                            if issubclass(item_attr_obj, commands.BaseCommand) and item_attr_obj is not commands.BaseCommand:
                                self.command_map[command_string] = item_attr_obj
    
    def _verify_command(self):
        """Called in __init__ to verify a submitted command is valid before running SkilletLoader"""
        if not self.action in self.command_map:
            print('Invalid action')
            exit(1)

    def _load_skillets(self):
        """Called in __init__ to front end SkilletLoader"""
        self.skillets = self.sl.load_all_skillets_from_dir(self.options.get('directory', './'))
        if len(self.sl.skillet_errors) > 0:
            print('Errors on loading skillets:')
            for err in self.sl.skillet_errors:
                for key in err:
                    print(f"   {key} - {err[key]}")

    def _verify_loaded_skillets(self):
        """Perform any pre-execution validations against loaded skillets"""

        if len(self.skillets) < 1:
            print("No skillets were loaded.")
            exit(1)

    def execute(self):
        """Run supplied SLI command"""

        action_obj = self.command_map[self.action](self)
        action_obj.execute()

        # Clean run, normal exit
        exit(0)


import yaml
import os
from panforge import Report

def print_table(objs, defs):
    """
    Print a formatted table with left justified columns
    """

    def get_cell(text, max):
        if text is None:
            return " " * max
        return f"  {text + ' ' * (max - len(text))}  "

    # Calculate max width of each column
    cols = {k: {"width": len(k)} for k in defs}
    for o in objs:
        for k in defs:
            if "." in defs[k]:
                text = jmespath.search(defs[k], o)
            else:
                text = o[defs[k]]
            if text is None:
                continue
            if len(text) > cols[k]["width"]:
                cols[k]["width"] = len(text)
    # Print headers
    header = "".join([get_cell(x, cols[x]["width"]) for x in cols])
    print(header)
    print("-" * len(header))

    # Print data
    for o in objs:
        row = ""
        for k in defs:
            if "." in defs[k]:
                text = jmespath.search(defs[k], o)
            else:
                text = o[defs[k]]
            row += get_cell(text, cols[k]["width"])
        print(row)

def load_config_file(fileName):
    """
    Attempt to load configuration file, return empty dict if not specified
    """
    if not fileName:
        return {}
    if not os.path.exists(fileName):
        print('Specified configuration file not found.')
        exit(1)
    with open(fileName, 'r') as f:
        return yaml.safe_load(f)

def generate_report(out_file, data, report_dir, header=None):
    """
    Generate a panforge report at 'out_file' from source 'data' using 'report_dir'
    as the root reporting directory for panforge
    """
    if not os.path.exists(report_dir):
        print(f'Could not generate report, source directory {report_dir} not present')
        return
    report = Report(report_dir)
    if header:
        report.load_header(header)
    report.load_data(data)
    report_html = report.render_html()
    with open(out_file, 'w') as f:
        f.write(report.html)
    print(f'Report written to {out_file}')

def input_yes_no(prompt, ret_value, ret_false=None):
    """
    Simple wrapper for asking user yes no questions until an acceptable answer
    is received, returns ret_value if yes, ret_false if no
    """
    while True:
        user_input = input(prompt + ' (y/n): ')
        if user_input.lower() in ('y', 'yes'):
            return ret_value
        if user_input.lower() in ('n', 'no'):
            return ret_false
        print('Please input either y or n')

def get_variable_input(var):
    """
    Get input from user in reference to var, being a skillet variable object
    """
    type_hint = var.get('type_hint')
    name = var.get('name')
    description = var.get('description')
    if not name:
        raise ValueError(f'Input variable missing name')
    ret_dict = {}

    if type_hint == 'checkbox':
        cbx_list = var.get('cbx_list', [])
        desc = description if description else name
        print(f"\n{desc}\n{'-'*len(desc)}\n")

        # Get inputs from user until user confirms they are ok
        confirmed = False
        input_list = []
        while not confirmed:
            input_list.clear()
            for cbx in cbx_list:
                input_value = input_yes_no(cbx['key'], cbx['value'])
                if input_value:
                    input_list.append(input_value)

            # Display all inputs
            check = []
            for cbx in cbx_list:
                check.append({'name':cbx['key'], 'value': 'Yes' if cbx['value'] in input_list else 'No'})
            print('')
            print_table(check, {'Input': 'name', 'Value': 'value'})
            print('')
            if input_yes_no('Are These answers ok?', True):
                confirmed = True

        # Put inputs inside ret_dict and return
        ret_dict[name] = input_list

        return ret_dict

    raise ValueError(f'Unsupported input variable type of {type_hint} in var {name}')


import click
from sli.skilletLineInterface import SkilletLineInterface
from sli.tools import load_config_file

@click.command()
@click.option("-c", "--config", help="Configuration file")
@click.option("-cm", "--commit", is_flag=True, help="Commit configuration changes")
@click.option("-v", "--verbose", is_flag=True, help="Verbose output")
@click.option("-d", "--device", help="Device IP or hostname")
@click.option("-sd", "--directory", help="Directory to load skillets from", default="./")
@click.option("-u", "--username", help="Device username")
@click.option("-p", "--password", help="Device password")
@click.option("-n", "--name", help="Name of skillet to execute")
@click.option("-r", "--report", is_flag=True, help="Generate a panforge formatted report")
@click.option("-rf", "--report-file", help="location of HTML file to create")
@click.argument("action", nargs=1, default="execute")
def cli(action, **kwargs):

    # Load configuration options from file, override with click parameters
    config_obj = load_config_file(kwargs.get('config'))
    for key in kwargs:
        if kwargs[key] is not None:
            config_obj[key] = kwargs[key]

    # Instantiate skillet and execute command
    sli = SkilletLineInterface(config_obj, action)
    sli.execute()


if __name__ == "__main__":
    cli()


import os
from .base import BaseCommand
from sli.tools import print_table, generate_report
from sli.decorators import (
    require_ngfw_connection_params,
    require_single_skillet,
    require_skillet_type
)

"""
Execute a validation skillet and generate results
"""

class ValidateCommand(BaseCommand):

    sli_command = 'validate'

    @require_single_skillet
    @require_skillet_type('pan_validation')
    @require_ngfw_connection_params
    def run(self):
        """
        SLI action, test load all skillets in directory and print out loaded skillets
        Generate panforge report if requested
        """

        exe = self.sli.skillet.execute(self.sli.context) 

        if self.sli.verbose:
            self.print_verbose(exe)
        self.print_summary(exe)

        if self.sli.generate_report:
            header = {'Host': self.sli.options['device']}
            out_file = getattr(self.sli, 'report_file', '')
            if len(out_file) < 1:
                out_file = f'{self.sli.options["device"]}-{self.sli.skillet.name}.html'
            report_dir = os.path.sep.join([self.sli.skillet.path, 'report'])
            generate_report(out_file, exe['pan_validation'], report_dir, header=header)

    def print_verbose(self, exe):
        print('Validation details\n------------------')
        for snippet_name in exe['pan_validation']:
            snippet = exe['pan_validation'][snippet_name]
            print(snippet_name)
            print('-'*len(snippet_name))
            if 'results' in snippet:
                result = "Passed" if snippet['results'] else "Failed"
                print(f"   Validation Results: {result}")
            print(f"   Label: {snippet['label']}")
            print(f"   Output: {snippet['output_message']}")
            documentation_link = snippet.get('documentation_link')
            if documentation_link:
                print(f"   Documentation Link: {documentation_link}")
            if isinstance(snippet['meta'], dict):
                print('   Meta:')
                for key in snippet['meta']:
                    print(f"      {key}: {snippet['meta'][key]}")
            print()
        

    def print_summary(self, exe):
        results = [
            {
                'name':x,
                'result': "Passed" if exe['snippets'][x] else "Failed"
            }
            for x in exe['snippets']
        ]
        print('Validation results\n------------------')
        print_table(results, {
            "Name": "name",
            "Result": "result"
        })
